#include "GEM.h"
#include "TMath.h"
#include <cmath>

ClassImp(TPC::GEM)
namespace TPC {
// C-tor
GEM::GEM(double sgprf, // PRF width      [mm]
         double wpad,  // pad pitch      [mm]
         double lpad,  // pad row pitch  [mm]
         double vgap,  // drift velocity [um/ns]
         double igap,  // induction gap  [mm]
         double gain,  // average gas gain
         double theta, // Polya parameter
	 bool   isodd) // (true,false)=(odd row, even row)
   : TF2   ("gem",
             this,
	    &TPC::GEM::Prf,      // member function pointer
            -5.*sgprf, 5.*sgprf, // (xmin, xmax) : avalanche arrival x range
            -5.*sgprf, 5.*sgprf, // (ymin, ymax) : avalanche arrival y range
             5,                  // # parameters (sgprf,wpad,lpad,vgap,igap)
            "TPC::GEM",          // class name
            "Prf"),              // member function name
     fVdrift(vgap),
     fIndGap(igap),
     fGain  (gain),
     fPolya (theta),
     fIsOdd (isodd)
{
    SetParameter(0, sgprf);
    SetParameter(1, wpad);
    SetParameter(2, lpad);
}

// 2D charge distribution for an avalanche at the origin
double GEM::Prf(const double *xv, const double *pv)
{
    double sg = pv[0]; // width of pad response function  [mm]
    double w  = pv[1]; // pad pitch                       [mm]
    double l  = pv[2]; // pad row pitch                   [mm]
    //double vg = pv[3]; // drift velocity in induction gap [um/ns]
    //double ig = pv[4]; // induction gap length            [mm]

    double x   = xv[0];   // electron arrival x position
    double y   = xv[1];   // electron arrival y position

    double val = 1./(2*TMath::Pi()*sg*sg)*exp(-(x*x+y*y)/(2.*sg*sg));
    return val;
}

// Charge fraction in a-th pad.
double GEM::Fx(int a, double x, double y)
{
    double sg = GetParameter(0); // sigma_PRF  [mm]
    double w  = GetParameter(1); // pad pitch  [mm]
    double l  = GetParameter(2); // pad row pitch  [mm]

    double xmin = w*(a - 0.5) - x + GetOffset();
    double xmax = w*(a + 0.5) - x + GetOffset();
    double ymin = -l/2 - y;
    double ymax = +l/2 - y;

    double val = this->Integral(xmin,xmax,ymin,ymax);
    return val;
}

double GEM::GasGain()
{
    return fGain*fPolya.GetRandom();
}

// PRF earea in a rectangular region: (xlo,xhi)x(ylo,yhi).
double GEM::Integral(double xlo, double xhi, double ylo, double yhi, double eps)
{
    double sg   = GetParameter(0); // sigma_PRF  [mm]
    double val  = (TMath::Erf((xhi)/sg/TMath::Sqrt2())
	          -TMath::Erf((xlo)/sg/TMath::Sqrt2()))/2;
#ifdef CHARGE_SHARING
           val *= (TMath::Erf((yhi)/sg/TMath::Sqrt2())
	          -TMath::Erf((ylo)/sg/TMath::Sqrt2()))/2;
#else
    if (ylo >= 0. || yhi <= 0.) val = 0.;
#endif

   return val;
}
}
