#include "Pasa.h"
#include "TMath.h"
#include <cmath>

ClassImp(TPC::Pasa)
namespace TPC {
// C-tor
Pasa::Pasa(double tpk, int n, double w, double noise, double norm)
   : TF2   ("pasa",
             this,
	    &TPC::Pasa::Charge, // member function pointer
            -100., 100., // (tmin, tmax) : avalanche arrival time range
             -20.,  20., // (bmin, bmax) : time slot range
                      5, // # parameters (tpk, n, w, noise, norm) 
            "TPC::Pasa", // class name
	    "Charge")    // member function name
{
    SetParameter(0, tpk);
    SetParameter(1, n);
    SetParameter(2, w);
    SetParameter(3, noise);
    SetParameter(4, norm);
}

// Charge fraction in a-th time slot.
double Pasa::Charge(const double *xv, const double *pv)
{
    double tpk = pv[0]; // shaper peaking time
    int    n   = pv[1]; // shaper order
    double w   = pv[2]; // time bin width
    double t   = xv[0]; // input signal arrival time
    double a   = xv[1]; // time slot number
    double tau = tpk/n; // shaper decay time
    double val = Area((a-0.5)*w-t, (a+0.5)*w-t, tau, n);
    return val;
}

// Charge fraction in a-th time slot.
double Pasa::Ft(double t, int a)
{
    double tpk = GetParameter(0); // shaper peaking time
    int    n   = GetParameter(1); // shaper order
    double w   = GetParameter(2); // time bin width
    double tau = tpk/n; // shaper decay time
    double val = Area((a-0.5)*w-t, (a+0.5)*w-t, tau, n);
    return val;
}

// Output pulse shape.
double Pasa::Pulse(const double *xv, const double *pv)
//double Pasa::Pulse(double *xv, double *pv)
{
    double tpk  = pv[0]; // shaper peaking time
    int    n    = pv[1]; // shaper order
    double norm = pv[4]; // pulse charge
    double t    = xv[0]; // input signal arrival time
    double tau  = tpk/n; // shaper decay time
    return norm*Pulse(t, tau, n);
}

// Output pulse shape.
double Pasa::Pulse(double t, double tau, int n) const
{
    if (t<=0) return 0.; // Theta(t/tau)
    double x   = t/tau;
    return TMath::Power(x,n)*TMath::Exp(-x)/TMath::Factorial(n)/tau;
}

// Pulse area in (a,b).
double Pasa::Area(double a, double b, double tau, int n) const
{
    if (a < 0. && b < 0.) return 0.;
    double val = tau*(Pulse(a,tau,n) - Pulse(b,tau,n));
    if (n) {
        val += Area(a,b,tau,n-1);
    } else if (a < 0.) {
        val += 1.; // contribution of dTheta(x)/dx=delta(x)
    }
    return val;
}
}
