#include "Track.h"

#include "GEN/Track.h"
#include "GEN/Cluster.h"
#include "GEN/Electron.h"

#include "TMath.h"
#include "TRandom.h"
#include <iostream>
#include <algorithm>
#if 1
#include "TFile.h"
#include "TROOT.h"
#endif

ClassImp(TPC::Track)
ClassImp(TPC::Track::Cluster)
ClassImp(TPC::Track::Cluster::SI)

namespace TPC {
// C-tor
Track::Track(double phi, // phi angle
             double lam, // dip angle
             double x0,  // x position [mm]
             double y0,  // y position [mm]
             double z0,  // z position [mm]
	     double l,   // projected track length [mm]
             double nav, // <# primary ionization/[cm]>
             double vd,  // drift veclocity [um/ns]
             double cdt, // Cd_T [um/sqrt{cm}]
             double cdl) // Cd_L [um/sqrt{cm}]
     : fPhi(phi),
       fLam(lam),
       fXv0(x0,y0,z0),
       fL(l),
       fNav(nav),
       fVdrift(vd),
       fCdT(cdt),
       fCdL(cdl)
{
    double csf = TMath::Cos(phi);
    double snf = TMath::Sin(phi);
    double tnl = TMath::Tan(lam);
    TVector3 ev (snf,csf,tnl); // track direction 

    double path   = l/csf*TMath::Sqrt(1+tnl*tnl); // path length
    double mu     = nav*path/10.;                 // <# primary ionizations>
    int    n      = gRandom->Poisson(mu);         // # primary ionizations
    // std::cerr << "mu=" << mu << " n=" << n  << std::endl;
    //--
    // Generate clusters.
    //--
    for (int i=0; i<n; i++) {
        TVector3 xv = fXv0 + path*(gRandom->Uniform()-0.5)*ev;
	// xv.Print();
        Add(new Cluster(xv,*this)); 
    }
    SetOwner(); // "Track" owns clusters
}

Track::Track(GEN::Track *gtkp,   // Heed track pointer
             double      vd,  // drift veclocity [um/ns]
             double      cdt, // Cd_T [um/sqrt{cm}]
             double      cdl) // Cd_L [um/sqrt{cm}]
     : fPhi(atan2(gtkp->GetP0().X(),gtkp->GetP0().Y())),
       fLam(atan2(gtkp->GetP0().Z(),gtkp->GetP0().Y())),
       fXv0(gtkp->GetX0().X(),gtkp->GetX0().Y(),gtkp->GetX0().Z()),
       fL(gtkp->GetTrackLength()),
       fNav(-999999),
       fVdrift(vd),
       fCdT(cdt),
       fCdL(cdl)
{
    fXv0 *= 10.; // [cm] -> [mm]
    fL   *= 10.; // [cm] -> [mm]
    //--
    // Copy Heed Track clusters.
    //--
    TIter next(gtkp);
    GEN::Cluster *gcp = nullptr;
    while ((gcp = static_cast<GEN::Cluster *>(next()))) { 
        Add(new Cluster(gcp,*this)); 
    }
    SetOwner(); // "Track" owns clusters
}

void Track::Shift(TVector3 dxv)
{
    fXv0 += dxv;
    TIter next(this);
    Cluster *clp;
    while ((clp = static_cast<Cluster *>(next()))) {
        clp->Shift(dxv);
    }
}

void Track::Drift(double d)
{
    TIter next(this);
    Cluster *clp;
    while ((clp = static_cast<Cluster *>(next()))) {
        clp->Drift(d);
    }
}

//-----------------------------------------------------------------------------
// Sub-classes
//-----------------------------------------------------------------------------
Track::Cluster::SI Track::Cluster::gSI;

Track::Cluster::Cluster(TVector3 &xv, Track &sd, int m)
     : fXv0  (xv), // primary ionization location [mm]
       fTime (0.), // drift time [ns]
       fTrack(sd)  // parent
{
#if 1
    if (m <= 0) m = GetSI().GetRandom() + 1; // decide cluster size
#else
    if (m <= 0) m = GetSI().TH1D::GetRandom() + 1; // decide cluster size
#endif
    ////////////////////////////////////////////////////////////////
    // FIXME
    //--
    // Decide the geometric size of this cluster.
    //   Very crude treatment now.
    //--
#if 0
    // switch off delta ray treatment
    static const double kRmin = 10.e2; // range cut [mm]
#else
    // do delta ray treatment
    static const double kRmin = 10.e-3; // range cut [mm]
#endif
    static const double kWAr  = 26.e-3; // Ar work fun. [keV]
    double e     = m*kWAr;   // [keV] .. very crude estimate
    double range = Range(e); // [mm]
    ////////////////////////////////////////////////////////////////
    if (range <= kRmin) { // if range < kRmin ignore cluster size.
        for (int i=0; i<m; i++) {
            Add(new TVector3(xv)); 
	}
    } else { // Consider cluster size due to delta-ray.
        double psi    = TMath::Pi()*gRandom->Uniform();
	double zrange = range * TMath::Cos(psi);
#if 0
	std::cerr << "m=" << m << " e=" << e << " R=" << range 
                  << " Rz=" << zrange << std::endl;
#endif
        // Primary electron
        Add(new TVector3(xv.X(),xv.Y(),xv.Z()+zrange)); 
	// Secondary electrons
        for (int i=1; i<m; i++) {
	    //--
            // delta-ray
	    //   Assume that B-field is strong enough so that
	    //   the delta-lay only affects z.
            //--
	    double dz = zrange*gRandom->Uniform(); 
            Add(new TVector3(xv.X(),xv.Y(),xv.Z()+dz)); 
	}
    }
    SetOwner(); // Cluster owns individual electron positions.
}

Track::Cluster::Cluster(GEN::Cluster *gcp, Track &sd)
     : fXv0  (gcp->GetPosition()), // primary ionization location [mm]
       fTime (0.), // drift time [ns]
       fTrack(sd)  // parent
{
     fXv0 *= 10.; // [cm]->[mm]
     TIter next(gcp);
     GEN::Electron *gep = nullptr;
     while ((gep = static_cast<GEN::Electron *>(next()))) {
           Add(new TVector3(10.*(gep->GetX0())));
     }
}


void Track::Cluster::Shift(TVector3 dxv)
{
     fXv0 += dxv;
     TIter next(this);
     TVector3 *xvp;
     while ((xvp = static_cast<TVector3 *>(next()))) {
        TVector3 &xv = *xvp;
        xv += dxv;
    }
}

void Track::Cluster::Drift(double d)
{
     if (d < 0.) d = fXv0.Z(); // if d<0, do full drift.

     double cdt  = fTrack.fCdT/1000.;
     double cdl  = fTrack.fCdL/1000.;
     double sgt  = cdt*TMath::Sqrt(d/10.);   // [mm]
     double sgl  = cdl*TMath::Sqrt(d/10.);   // [mm]

     TIter next(this);
     TVector3 *xvp;
     while ((xvp = static_cast<TVector3 *>(next()))) {
        TVector3 &xv = *xvp;
        xv[0] += gRandom->Gaus(0.,sgt);
        xv[1] += gRandom->Gaus(0.,sgt);
        xv[2] += gRandom->Gaus(0.,sgl) - d;
	//xv.Print();
    }
    fTime = d*1000./fTrack.fVdrift; // [ns]
}

double Track::Cluster::Range(double e /* [keV]*/) const
{
    // delta-ray range [g/cm^2]
    static const double a = 5.37e-4;   // [g/cm^2/keV]
    static const double b = 0.9815;
    static const double c = 3.1230e-3; // [1/keV]
    double range = a*e*(1.-b/(1+c*e));
    // convert to mm
    static const double rho = 36./22.4e3; // Ar density [g/cm^3]
    range *= 10./rho; // [mm]
    return range;
}

int Track::Cluster::Compare(const TObject *obj) const
{
    const Track::Cluster *cp = static_cast<const Track::Cluster *>(obj);
    double yme  = fXv0.Y();
    double yyou = cp->GetXv0().Y();
    return  yme < yyou ? -1 : (yme > yyou ? +1 : 0);
}

Track::Cluster::SI::SI()
     : TH1D("hProbSI","",5000,0.,5000)
{
    static const double Ar[] = { 65.60, 15.00, 6.40, 3.50, 2.25,
                                  1.55,  1.05, 0.81, 0.61, 0.49,
                                  0.39,  0.30, 0.25, 0.20, 0.16,
                                  0.12,  0.095, 0.075, 0.063, 21.6 };
    const int kMax = 3000;
    double sum = 0;
    fPSI.reserve(kMax);
    for (int i=0; i<kMax; i++) {
#if 0
      double val;
      if (i<19) val = Ar[i];
      else      val = Ar[19]/(i+1)/(i+1);
#else
      static TH1D *hp = nullptr;
      if (!hp) {
         auto ncfilep = new TFile("Data/nc.electron.5GeV.root");
         hp = static_cast<TH1D *>(gROOT->FindObject("hnc"));
      }
#if 0 // bug fix: 2021/03/20
      double val = hp->GetBinContent(i);
#else
      double val = hp->GetBinContent(i+1);
#endif
#endif
      sum += val;
      fPSI.push_back(val);
    }

    for (int i=0; i<kMax; i++) {
        fPSI[i] /= sum;  // normalize to unity
	Fill(i,fPSI[i]); // fill up the PSI histogram.
    }
}

double Track::Cluster::SI::GetPSI(int m) const
{
    return fPSI[m];
}

double Track::Cluster::SI::GetRandom() const
{
    static std::vector<double> *cumPSIPtr = nullptr;
    if (!cumPSIPtr) {
        cumPSIPtr = new std::vector<double>;
	double sum = 0.;
	int    npt = 0;
	for (double x : fPSI) {
            sum += x;
	    cumPSIPtr->push_back(sum);
	    npt++;
	}
	(*cumPSIPtr)[npt-1] = 1.;
    }
    std::vector<double> &cumPSI = *cumPSIPtr;
    double r   = gRandom->Uniform();
    auto   loc = std::find_if(cumPSI.begin(), cumPSI.end(), [&](double y) { return y > r; });

    return distance(cumPSI.begin(), loc);;
}
}
