#ifndef TPC_TRACK_H
#define TPC_TRACK_H
#include "TVector3.h"
#include "TObjArray.h"
#include "TH1D.h"
//-----------------------------------------------------------------------------
// ===========
// class Track
// ===========
// (Description)
//    Classes to describe a track of ionization.
//    A "Track" is a set of "Cluster"s, while a "Cluster"
//    consists of a set of primary + secondary ionization electrons.
// (Provides)
//    TPC::Track              : track object
//    TPC::Track::Cluster     : cluster object
//    TPC::Track::Cluster::SI : 2ndary ionization handler
// (Update Record)
//    2015/07/10  K.Fujii     Original version.
//-----------------------------------------------------------------------------
namespace GEN {
class Track;
class Cluster;
}
namespace TPC {
class Track : public TObjArray {
public:
    Track(double phi =   0.,  // phi angle
          double lam =   0.,  // dip angle
          double x0  =   0.,  // x position [mm]
          double y0  =   0.,  // y position [mm]
          double z0  = 100.,  // z position [mm]
          double l   =  10.,  // projected track length [mm]
          double nav =  30.,  // <# primary ionization/[cm]>
          double vd  =  75.,  // drift veclocity [um/ns]
          double cdt =  95.,  // Cd_T [um/sqrt{cm}]
          double cdl = 220.); // Cd_L [um/sqrt{cm}]

    Track(GEN::Track *gtkp,        // Heed track pointer
          double      vd  =  75.,  // drift veclocity [um/ns]
          double      cdt =  95.,  // Cd_T [um/sqrt{cm}]
          double      cdl = 220.); // Cd_L [um/sqrt{cm}]

    inline int    GetNav   () const { return fNav;    }
    inline double GetVdrift() const { return fVdrift; }
    inline double GetCdT   () const { return fCdT;    }
    inline double GetCdL   () const { return fCdL;    }

    inline const TVector3 &GetXv0() const { return fXv0;   }

    void Shift(TVector3 dxv); // Shift Track by dxv
    void Drift(double d = -1.);

    class  Cluster;
    friend Cluster;

    class Cluster : public TObjArray {
    public:
        Cluster(TVector3 &xv, Track &sd, int m = 0);
        Cluster(GEN::Cluster *gcp, Track &sd);

	class SI;
	inline const TVector3 &GetXv0   () const { return fXv0;   }
	inline const Track    &GetParent() const { return fTrack; }
	inline const SI       &GetSI    () const { return gSI;    }
	inline       double    GetTime  () const { return fTime;  }

                     int       Compare   (const TObject *obj) const;
                     bool      IsSortable() const { return kTRUE; }

        void     Shift(TVector3 dxv); // Shift Track by dxv
        void     Drift(double d = -1.);

	double   Range(double e) const;

	class SI : public TH1D {
        public:
            SI();

	    double GetRandom() const;
	    double GetPSI   (int m) const;

        private:
            std::vector<double> fPSI;  // P_SI[m]

            ClassDef(SI, 1)   // 2ndary ionization class
        };

    private:
        TVector3   fXv0;     // primary ionization point
	double     fTime;    // drift time [ns]
        Track     &fTrack;   // seeds
	static SI  gSI;      // 2ndary ionization handler

        ClassDef(Cluster, 1)   // Cluster class
    };

private:
    double   fPhi;    // phi angle
    double   fLam;    // dip angle
    TVector3 fXv0;    // track center position [mm]
    double   fL;      // projected track length [mm]
    double   fNav;    // <#primary ionizations>
    double   fVdrift; // drift velocity [um/ns]
    double   fCdT;    // CdT [um/sqrt{cm}]
    double   fCdL;    // CdL [um/sqrt{cm}]

    ClassDef(Track, 1)   // Track class
};
}
#endif
