//////////////////////
#define USER_TUPLE
#define STAGGERED_PADS
//////////////////////
#ifndef __CINT__
#include "TSystem.h"
#include "TMath.h"
#include "TFile.h"
#include "TNtupleD.h"
#include "TRandom.h"
#include "TH1D.h"
#include "TROOT.h"

#include <fstream>
#include <sstream>
#include <iomanip>

#include "GEM.h"
#include "Pasa.h"
#include "Polya.h"
#include "Hit.h"
#include "Track.h"
#endif

using namespace std;

//
// Track fitting version to see hodoscope effects
//
//

void Fit(int, double, double *, double *, double &, double &, int);

void ResolFit(double z0    = 0.00,   // drift distance [mm]
              double phi0  = 0.00,   // track phi      [rad]
              double dphi0 = 0.01,   // track sgphi    [rad]
              double lam0  = 0.00,   // track lambda   [rad]
              double dlam0 = 0.01,   // track sglambda [rad]
	      int    nev   = 40000, // # events to generate
	      bool   isCOG = true)  // time calculation method
{ 
    //--
    // PASA
    //--
    const double tpk0  = 120.; // pekaing time   [ns]
    const int    n0    = 4;    // shaper order
    const double w0    = 50.;  // time bin width [ns]
#if 0
    const double noise = 2.;  // noise count
    const double norm  = 0.014; // conversion factor: #e's to adc
#else
    //const double noise = 1.;  // noise count
    const double noise = 2.;  // noise count
    const double norm  = 0.0061; // conversion factor: #e's to adc
#endif

    TPC::Pasa  pca(tpk0,n0,w0,noise,norm);

    //--
    // GEM
    //--
#if 1
    double sgprf = 0.27; // sigma_PRF  [mm]
#else
    double sgprf = 0.22; // sigma_PRF  [mm]
#endif
    double wpad  = 1.15; // pad pitch  [mm]
    double lpad  = 5.36; // pad length [mm]
    double len   = 10.;  // seg length [mm]
#if 0
    double nav   = 30.;  // <#primary ionizations>
           nav  *= 1.2;  // 5 GeV electron > MIP
#else
#if 1
  double nav   = 27.5;  // <#primary ionizations> [1/cm] for MIP as from HEED (T2K)
         nav  *= 1.38;  // 5 GeV electron > MIP
#else
  double nav   = 15.5;  // <#primary ionizations> [1/cm] for MIP
         nav  *= 1.38;  // 5 GeV electron > MIP
#endif
#endif
    double vd    = 75.00; // drift velocity [um/ns]
    double vgap  = 50.00; // drift velocity [um/ns]
    double igap  =  2.00; // induction gap  [mm]
    double gain  = 2000.; // average gas gain
    double theta =  0.50; // polya parameter

    const int kNrows =  28;  // #pad rows

    TPC::GEM gem(sgprf,wpad,lpad,vgap,igap,gain,theta);

    double wzee  = vd*w0/1000.;// time bin width [mm]

    //--
    // Calculate position resolutions.
    //--
    TPC::Hit::Pulse::VGetTimeMethod *methodp = 0;
    if (isCOG) methodp = new TPC::Hit::Pulse::Cog(&pca);
    else       methodp = new TPC::Hit::Pulse::Inflection(&pca);

    TPC::Hit::Pulse::SetGetTimeMethodPtr(methodp);

    int  npad = 1; // use pads in (-n,+n) for time calculation
    //int  npad = 1; // use pads in (-n,+n) for time calculation
    if (isCOG) npad = 3;

    //--
    // Z loop parameters.
    //--
#if 0
    const int kNz = 24;  // # points 
    double zstep  = 25.; // [mm]
#else
    const int kNz = 12;  // # points 
    double zstep  = 50.; // [mm]
#endif

    //--
    //  Open root file.
    //--
    stringstream fname;
    if (isCOG) fname << "tupres.dray.cog";
    else       fname << "tupres.dray.inf";
    fname << ".phi"  << setw(4) << setprecision(2) << setiosflags(ios::fixed) << phi0
          << ".lam"  << setw(4) << setprecision(2) << setiosflags(ios::fixed) << lam0
          << ".z"    << setw(3) << setprecision(0) << setiosflags(ios::fixed) 
                                                 << setfill('0')            << z0
          << ".nav"  << setw(4) << setprecision(1) << setiosflags(ios::fixed) << nav
          << ".sgpr" << setw(4) << setprecision(2) << setiosflags(ios::fixed) << sgprf
          << ".tbin" << setw(2) << setprecision(0) << setiosflags(ios::fixed) 
                                                 << setfill('0')            << w0
          << ".tpk"  << setw(3) << setprecision(0) << setiosflags(ios::fixed)
                                                 << setfill('0')            << tpk0
          << ".nois" << setw(3) << setprecision(0) << setiosflags(ios::fixed)
                                                 << setfill('0')            << noise
          << ".root" << ends;
    auto filep = new TFile(fname.str().data(),"RECREATE");
    filep->cd();

    //--
    // Define residual ntuples.
    //--
    std::stringstream sout;
    sout << "nev:row:ipad:itbn:xt:yt:zt:q:dxin:dzin:dxot:dzot"
         << ends;
    auto tup = new TNtupleD("tupres","",sout.str().data());

    //--
    // Loop over events.
    //--
    const int kNev = nev;
    for (int iev=0; iev<kNev; iev++) {
#if 1
        cerr << "Event: " << iev << endl;
#endif
        double x0  = gRandom->Uniform(-wpad/2,+wpad/2); // pad scan
        double y0  = 0.;
        double dz0 = gRandom->Uniform(-wzee/2,+wzee/2); // time bin scan
        // FIXME
        //   do phi and lam scan here.
        double phi  = phi0 + dphi0*gRandom->Gaus();
        double lam  = lam0 + dlam0*gRandom->Gaus();
        double sinf = TMath::Sin(phi);
        double tanl = TMath::Tan(lam);
        //--
        // Loop over pad rows and create hits.
        //--
        double xreco[kNrows]; // reconstructed x
        double zreco[kNrows]; // reconstructed z
        double yrow [kNrows]; // y position of each pad row
        double qreco[kNrows]; // charge
        int    ilc = kNrows/2; // center rows.
        for (int il=0; il<kNrows; il++) { 
            double yt0  = y0 + (il-ilc)*lpad;
            double xt0  = x0 + (yt0-y0)*sinf;
            double zt0  = z0 + (yt0-y0)*tanl;
            double fxt0 = xt0/wpad;
            int    ipad = (xt0 + (fxt0 > 0.5 ? 0.5 : (fxt0 < -0.5 ? -0.5 : 0.0))*wpad)/wpad; // center pad #
            double fzt0 = zt0/wzee;
            int    itbn = (zt0 + (fzt0 > 0.5 ? 0.5 : (fzt0 < -0.5 ? -0.5 : 0.0))*wzee)/wzee; // time bin #
            double xt   = xt0 - ipad*wpad;       // x measured from pad center
            double zt   = zt0 - itbn*wzee;       // z measured from t0
               yrow[il] = yt0;
         
            TPC::Track trk(phi,lam,xt,y0,zt0+dz0,len,nav,vd); // create a track segment
            trk.Drift(zt0-zt);                                // drift it to GEM

#ifdef STAGGERED_PADS
            gem.SetIsOdd(il%2 ? true : false);                // stagger odd rows
#endif
            TPC::Hit hit(trk, gem, pca);                      // create signals
            xreco[il] = hit.GetX() + ipad*wpad;               // reconstructed x
           
            if (isCOG) zreco[il] = hit.GetZ(npad) + itbn*wzee - 2.2; // shift due to asymmetry (cog)
            else       zreco[il] = hit.GetZ(npad) + itbn*wzee + 2.2; // shift due to asymmetry (inf)

            qreco[il] = hit.GetQ();
            //cerr << "il=" << il << " xreco=" << xreco[il] << " zreco=" << zreco[il] << endl;
        } // end loop over pad rows
        //--
        // Do track fitting.
        //   Ignore resolutions dependence on pad rows.
        //--
        double x0in;
        double dxdyin;
        Fit(kNrows,y0,xreco,yrow,x0in,dxdyin,-99999);  // including ilc
 
        double x0ot;
        double dxdyot;
        Fit(kNrows,y0,xreco,yrow,x0ot,dxdyot,ilc);     // excluding ilc
 
        double z0in;
        double dzdyin;
        Fit(kNrows,y0,zreco,yrow,z0in,dzdyin,-99999);  // including ilc
 
        double z0ot;
        double dzdyot;
        Fit(kNrows,y0,zreco,yrow,z0ot,dzdyot,ilc);     // excluding ilc
 
        //--
        // Calculate residuals, etc.
        //--
        double yt0  = yrow[ilc];
        double xt0in  = x0in + (yt0-y0)*dxdyin;
        double zt0in  = z0in + (yt0-y0)*dzdyin;
        double xt0ot  = x0ot + (yt0-y0)*dxdyot;
        double zt0ot  = z0ot + (yt0-y0)*dzdyot;
 
        double delxin = xreco[ilc] - xt0in;
        double delzin = zreco[ilc] - zt0in;
        double delxot = xreco[ilc] - xt0ot;
        double delzot = zreco[ilc] - zt0ot;
 
        int    itbn   = (z0 - 0.5*wzee)/wzee; // time bin #
 
        int    ipad   = 0; // pad #
        tup->Fill(iev,ilc,ipad,itbn,x0,y0,dz0,qreco[ilc],delxin,delzin,delxot,delzot);

#if 0
            cerr << "iev="   << iev    << " ilc= " << ilc << " q=" << qreco[ilc]
                 << " dxin=" << delxin << " dzin=" << delzin
                 << " dxot=" << delxot << " dzot=" << delzot
                 << endl;
#endif

    } // end loop over events.
    filep->Write();
    //TPC::Hit::ResetTup();

#if 1
    //--
    // x resolution
    //--
    tup->Draw("dxin>>hxin(200,-1.,+1.)");
    auto hxin = static_cast<TH1D *>(gROOT->FindObject("hxin"));
    double  sgxin = hxin->GetRMS();
    double dsgxin = hxin->GetRMSError();
    
    tup->Draw("dxot>>hxot(200,-1.,+1.)");
    auto hxot = static_cast<TH1D *>(gROOT->FindObject("hxot"));
    double  sgxot = hxot->GetRMS();
    double dsgxot = hxot->GetRMSError();
    
    double  sgx = TMath::Sqrt(sgxin*sgxot);
    double dsgx = (dsgxin+dsgxot)/2;
    
    //--
    // z resolution
    //--
    tup->Draw("dzin>>hzin(200,-2.,+2.)");
    auto hzin = static_cast<TH1D *>(gROOT->FindObject("hzin"));
    double  sgzin = hzin->GetRMS();
    double dsgzin = hzin->GetRMSError();
    
    tup->Draw("dzot>>hzot(200,-2.,+2.)");
    auto hzot = static_cast<TH1D *>(gROOT->FindObject("hzot"));
    double  sgzot = hzot->GetRMS();
    double dsgzot = hzot->GetRMSError();
    
    double  sgz = TMath::Sqrt(sgzin*sgzot);
    double dsgz = (dsgzin+dsgzot)/2;
    
  //--
  // Output residual data.
  //--
    stringstream sgout;
    if (isCOG) sgout << "sg.cog.dat" << ends;
    else       sgout << "sg.inf.dat" << ends;
    ofstream fout(sgout.str().data(),ios::app);
    fout << setw(5) << setprecision(3) << setiosflags(ios::fixed) << phi0 << " "
         << setw(5) << setprecision(3) << setiosflags(ios::fixed) << lam0 << " "
         << setw(4) << setprecision(0) << setiosflags(ios::fixed) << z0   << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << sgx  << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << dsgx << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << sgz  << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << dsgz
         << endl;
  
    cerr << setw(5) << setprecision(3) << setiosflags(ios::fixed) << phi0 << " "
         << setw(5) << setprecision(3) << setiosflags(ios::fixed) << lam0 << " "
         << setw(4) << setprecision(0) << setiosflags(ios::fixed) << z0   << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << sgx  << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << dsgx << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << sgz  << " "
         << setw(8) << setprecision(6) << setiosflags(ios::fixed) << dsgz
         << endl;
#endif
}

void Fit(int nrows, double y0, double *xreco, double *yrow, double &x0, double &dxdy, int ix)
{
    double a11 = 0.;
    double a12 = 0.;
    double a22 = 0.;
    double b1  = 0.;
    double b2  = 0.;
    for (int i=0; i<nrows; i++) {
        if (i == ix) continue;
        a11 += 1.;
        double dy   = yrow[i] - y0;
        a12 += dy;
        a22 += dy*dy;
        b1  += xreco[i];
        b2  += xreco[i]*dy;
    }
    double det = a11*a22 - a12*a12;
    double ainv11 =  a22/det;
    double ainv12 = -a12/det;
    double ainv21 =  ainv12;
    double ainv22 =  a11/det;

    x0   = ainv11*b1 + ainv12*b2;
    dxdy = ainv21*b1 + ainv22*b2;
}
